@@ -27,13 +27,14 @@ Released under the MIT licence.
 # DEALINGS IN THE SOFTWARE.
 
 __author__ = "Marius Gedminas (marius@gedmin.as)"
-__copyright__ = "Copyright (c) 2008-2010 Marius Gedminas"
+__copyright__ = "Copyright (c) 2008-2011 Marius Gedminas"
 __license__ = "MIT"
-__version__ = "1.6.0"
-__date__ = "2010-12-18"
+__version__ = "1.7.0"
+__date__ = "2011-03-11"
 
 
 import gc
+import re
 import inspect
 import types
 import operator
@@ -57,22 +58,28 @@ except AttributeError:
     iteritems = dict.items
 
 
-def count(typename):
+def count(typename, objects=None):
     """Count objects tracked by the garbage collector with a given class name.
 
     Example:
 
         >>> count('dict')
         42
-        >>> count('MyClass')
+        >>> count('MyClass', get_leaking_objects())
         3
 
     Note that the GC does not track simple objects like int or str.
+
+    .. versionchanged:: 1.7
+       New parameter: ``objects``.
+
     """
-    return sum(1 for o in gc.get_objects() if type(o).__name__ == typename)
+    if objects is None:
+        objects = gc.get_objects()
+    return sum(1 for o in objects if type(o).__name__ == typename)
 
 
-def typestats():
+def typestats(objects=None):
     """Count the number of instances for each type tracked by the GC.
 
     Note that the GC does not track simple objects like int or str.
@@ -84,17 +91,25 @@ def typestats():
 
         >>> typestats()
         {'list': 12041, 'tuple': 10245, ...}
+        >>> typestats(get_leaking_objects())
+        {'MemoryError': 1, 'tuple': 2795, 'RuntimeError': 1, 'list': 47, ...}
 
     .. versionadded:: 1.1
+
+    .. versionchanged:: 1.7
+       New parameter: ``objects``.
+
     """
+    if objects is None:
+        objects = gc.get_objects()
     stats = {}
-    for o in gc.get_objects():
+    for o in objects:
         stats.setdefault(type(o).__name__, 0)
         stats[type(o).__name__] += 1
     return stats
 
 
-def most_common_types(limit=10):
+def most_common_types(limit=10, objects=None):
     """Count the names of types with the most instances.
 
     Returns a list of (type_name, count), sorted most-frequent-first.
@@ -110,16 +125,20 @@ def most_common_types(limit=10):
         [('list', 12041), ('tuple', 10245)]
 
     .. versionadded:: 1.4
+
+    .. versionchanged:: 1.7
+       New parameter: ``objects``.
+
     """
-    stats = sorted(typestats().items(), key=operator.itemgetter(1),
+    stats = sorted(typestats(objects).items(), key=operator.itemgetter(1),
                    reverse=True)
     if limit:
         stats = stats[:limit]
     return stats
 
 
-def show_most_common_types(limit=10):
-    """Print the table of types of most common instances
+def show_most_common_types(limit=10, objects=None):
+    """Print the table of types of most common instances.
 
     The caveats documented in :func:`typestats` apply.
 
@@ -133,8 +152,12 @@ def show_most_common_types(limit=10):
         builtin_function_or_method 800
 
     .. versionadded:: 1.1
+
+    .. versionchanged:: 1.7
+       New parameter: ``objects``.
+
     """
-    stats = most_common_types(limit)
+    stats = most_common_types(limit, objects)
     width = max(len(name) for name, count in stats)
     for name, count in stats:
         print('%-*s %i' % (width, name, count))
@@ -180,7 +203,30 @@ def show_growth(limit=10, peak_stats={}):
             print('%-*s%9d %+9d' % (width, name, stats[name], delta))
 
 
-def by_type(typename):
+def get_leaking_objects(objects=None):
+    """Return objects that do not have any referents.
+
+    These could indicate reference-counting bugs in C code.  Or they could
+    be legitimate.
+
+    Note that the GC does not track simple objects like int or str.
+
+    .. versionadded:: 1.7
+    """
+    if objects is None:
+        gc.collect()
+        objects = gc.get_objects()
+    try:
+        ids = set(id(i) for i in objects)
+        for i in objects:
+            ids.difference_update(id(j) for j in gc.get_referents(i))
+        # this then is our set of objects without referrers
+        return [i for i in objects if id(i) in ids]
+    finally:
+        objects = i = j = None # clear cyclic references to frame
+
+
+def by_type(typename, objects=None):
     """Return objects tracked by the garbage collector with a given class name.
 
     Example:
@@ -189,8 +235,14 @@ def by_type(typename):
         [<mymodule.MyClass object at 0x...>]
 
     Note that the GC does not track simple objects like int or str.
+
+    .. versionchanged:: 1.7
+       New parameter: ``objects``.
+
     """
-    return [o for o in gc.get_objects() if type(o).__name__ == typename]
+    if objects is None:
+        objects = gc.get_objects()
+    return [o for o in objects if type(o).__name__ == typename]
 
 
 def at(addr):
@@ -210,6 +262,31 @@ def at(addr):
     return None
 
 
+def find_ref_chain(obj, predicate, max_depth=20, extra_ignore=()):
+    """Find a shortest chain of references leading from obj.
+
+    The end of the chain will be some object that matches your predicate.
+
+    ``predicate`` is a function taking one argument and returning a boolean.
+
+    ``max_depth`` limits the search depth.
+
+    ``extra_ignore`` can be a list of object IDs to exclude those objects from
+    your search.
+
+    Example:
+
+        >>> find_chain(obj, lambda x: isinstance(x, MyClass))
+        [obj, ..., <MyClass object at ...>]
+
+    Returns ``[obj]`` if such a chain could not be found.
+
+    .. versionadded:: 1.7
+    """
+    return find_chain(obj, predicate, gc.get_referents,
+                      max_depth=max_depth, extra_ignore=extra_ignore)[::-1]
+
+
 def find_backref_chain(obj, predicate, max_depth=20, extra_ignore=()):
     """Find a shortest chain of references leading to obj.
 
@@ -233,43 +310,14 @@ def find_backref_chain(obj, predicate, max_depth=20, extra_ignore=()):
        Returns ``obj`` instead of ``None`` when a chain could not be found.
 
     """
-    queue = [obj]
-    depth = {id(obj): 0}
-    parent = {id(obj): None}
-    ignore = set(extra_ignore)
-    ignore.add(id(extra_ignore))
-    ignore.add(id(queue))
-    ignore.add(id(depth))
-    ignore.add(id(parent))
-    ignore.add(id(ignore))
-    ignore.add(id(sys._getframe()))  # this function
-    gc.collect()
-    while queue:
-        target = queue.pop(0)
-        if predicate(target):
-            chain = [target]
-            while parent[id(target)] is not None:
-                target = parent[id(target)]
-                chain.append(target)
-            return chain
-        tdepth = depth[id(target)]
-        if tdepth < max_depth:
-            referrers = gc.get_referrers(target)
-            ignore.add(id(referrers))
-            for source in referrers:
-                if id(source) in ignore:
-                    continue
-                if id(source) not in depth:
-                    depth[id(source)] = tdepth + 1
-                    parent[id(source)] = target
-                    queue.append(source)
-    return [obj] # not found
+    return find_chain(obj, predicate, gc.get_referrers,
+                      max_depth=max_depth, extra_ignore=extra_ignore)
 
 
 def show_backrefs(objs, max_depth=3, extra_ignore=(), filter=None, too_many=10,
                   highlight=None, filename=None, extra_info=None,
                   refcounts=False):
-    """Generate an object reference graph ending at ``objs``
+    """Generate an object reference graph ending at ``objs``.
 
     The graph will show you what objects refer to ``objs``, directly and
     indirectly.
@@ -323,7 +371,7 @@ def show_backrefs(objs, max_depth=3, extra_ignore=(), filter=None, too_many=10,
 def show_refs(objs, max_depth=3, extra_ignore=(), filter=None, too_many=10,
               highlight=None, filename=None, extra_info=None,
               refcounts=False):
-    """Generate an object reference graph starting at ``objs``
+    """Generate an object reference graph starting at ``objs``.
 
     The graph will show you what objects are reachable from ``objs``, directly
     and indirectly.
@@ -378,25 +426,81 @@ def show_refs(objs, max_depth=3, extra_ignore=(), filter=None, too_many=10,
 def show_chain(*chains, **kw):
     """Show a chain (or several chains) of object references.
 
-    Useful in combination with :func:`find_backref_chain`, e.g.
+    Useful in combination with :func:`find_ref_chain` or
+    :func:`find_backref_chain`, e.g.
 
         >>> show_chain(find_backref_chain(obj, inspect.ismodule))
 
+    You can specify if you want that chain traced backwards or forwards
+    by passing a ``backrefs`` keyword argument, e.g.
+
+        >>> show_chain(find_ref_chain(obj, inspect.ismodule),
+        ...            backrefs=False)
+
+    Ideally this shouldn't matter, but for some objects
+    :func:`gc.get_referrers` and :func:`gc.get_referents` are not perfectly
+    symmetrical.
+
     You can specify ``highlight``, ``extra_info`` or ``filename`` arguments
-    like for :func:`show_backrefs`.
+    like for :func:`show_backrefs` or :func:`show_refs`.
 
     .. versionadded:: 1.5
+
+    .. versionchanged:: 1.7
+       New parameter: ``backrefs``.
+
     """
+    backrefs = kw.pop('backrefs', True)
     chains = [chain for chain in chains if chain] # remove empty ones
     def in_chains(x, ids=set(map(id, itertools.chain(*chains)))):
         return id(x) in ids
-    show_backrefs([chain[-1] for chain in chains], max(map(len, chains)) - 1,
+    max_depth = max(map(len, chains)) - 1
+    if backrefs:
+        show_backrefs([chain[-1] for chain in chains], max_depth=max_depth,
+                      filter=in_chains, **kw)
+    else:
+        show_refs([chain[0] for chain in chains], max_depth=max_depth,
                   filter=in_chains, **kw)
 
 #
 # Internal helpers
 #
 
+def find_chain(obj, predicate, edge_func, max_depth=20, extra_ignore=()):
+    queue = [obj]
+    depth = {id(obj): 0}
+    parent = {id(obj): None}
+    ignore = set(extra_ignore)
+    ignore.add(id(extra_ignore))
+    ignore.add(id(queue))
+    ignore.add(id(depth))
+    ignore.add(id(parent))
+    ignore.add(id(ignore))
+    ignore.add(id(sys._getframe()))  # this function
+    ignore.add(id(sys._getframe(1))) # find_chain/find_backref_chain, most likely
+    gc.collect()
+    while queue:
+        target = queue.pop(0)
+        if predicate(target):
+            chain = [target]
+            while parent[id(target)] is not None:
+                target = parent[id(target)]
+                chain.append(target)
+            return chain
+        tdepth = depth[id(target)]
+        if tdepth < max_depth:
+            referrers = edge_func(target)
+            ignore.add(id(referrers))
+            for source in referrers:
+                if id(source) in ignore:
+                    continue
+                if id(source) not in depth:
+                    depth[id(source)] = tdepth + 1
+                    parent[id(source)] = target
+                    queue.append(source)
+    return [obj] # not found
+
+
 def show_graph(objs, edge_func, swap_source_target,
                max_depth=3, extra_ignore=(), filter=None, too_many=10,
                highlight=None, filename=None, extra_info=None,
@@ -493,6 +597,9 @@ def show_graph(objs, edge_func, swap_source_target,
     f.write("}\n")
     f.close()
     print("Graph written to %s (%d nodes)" % (dot_filename, nodes))
+    if filename and filename.endswith('.dot'):
+        # nothing else to do, the user asked for a .dot file
+        return
     if not filename and program_in_path('xdot'):
         print("Spawning graph viewer (xdot)")
         subprocess.Popen(['xdot', dot_filename], close_fds=True)
@@ -539,7 +646,10 @@ def obj_label(obj, extra_info=None, refcounts=False):
 
 
 def quote(s):
-    return s.replace("\\", "\\\\").replace("\"", "\\\"").replace("\n", "\\n")
+    return (s.replace("\\", "\\\\")
+             .replace("\"", "\\\"")
+             .replace("\n", "\\n")
+             .replace("\0", "\\\\0"))
 
 
 def safe_repr(obj):
@@ -606,16 +716,24 @@ def edge_label(source, target):
                 return ' [label="im_self",weight=10]'
             if target is source.im_func:
                 return ' [label="im_func",weight=10]'
+    if isinstance(source, types.FunctionType):
+        for k in dir(source):
+            if target is getattr(source, k):
+                return ' [label="%s",weight=10]' % quote(k)
     if isinstance(source, dict):
         for k, v in iteritems(source):
             if v is target:
-                if isinstance(k, basestring) and k:
+                if isinstance(k, basestring) and is_identifier(k):
                     return ' [label="%s",weight=2]' % quote(k)
                 else:
-                    return ' [label="%s"]' % quote(safe_repr(k))
+                    return ' [label="%s"]' % quote(type(k).__name__ + "\n"
+                                                   + safe_repr(k))
     return ''
 
 
+is_identifier = re.compile('[a-zA-Z_][a-zA-Z_0-9]*$').match
+
+
 def program_in_path(program):
     path = os.environ.get("PATH", os.defpath).split(os.pathsep)
     path = [os.path.join(dir, program) for dir in path]
